package com.sap.imagineering.wellnessportal.service;

import java.math.BigDecimal;
import java.util.Set;

import javax.annotation.Resource;

import org.apache.log4j.Logger;
import org.scribe.model.Token;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.sap.imagineering.wellnessportal.api.exception.InternalServerException;
import com.sap.imagineering.wellnessportal.db.dao.MeasurementsDao;
import com.sap.imagineering.wellnessportal.db.dao.ProviderDao;
import com.sap.imagineering.wellnessportal.db.dao.UserDao;
import com.sap.imagineering.wellnessportal.db.domain.BadgeMeasurement;
import com.sap.imagineering.wellnessportal.db.domain.Measurement;
import com.sap.imagineering.wellnessportal.db.domain.ProviderCredentials;
import com.sap.imagineering.wellnessportal.db.domain.ProviderType;
import com.sap.imagineering.wellnessportal.db.domain.User;
import com.sap.imagineering.wellnessportal.provider.bodymedia.BMGoalTypeFactory;
import com.sap.imagineering.wellnessportal.provider.bodymedia.dto.BMBaseMeasurementDto;
import com.sap.imagineering.wellnessportal.provider.bodymedia.dto.BMGoal;
import com.sap.imagineering.wellnessportal.provider.bodymedia.dto.BMGoalsDto;
import com.sap.imagineering.wellnessportal.provider.bodymedia.dto.BMMeasurements;
import com.sap.imagineering.wellnessportal.provider.bodymedia.dto.BMWeightMeasureDto;


@Service("bodymediaDailySynch")
public class BodymediaDailySynch extends AbstractDailySynch {
	
	
	private static final Logger LOG = Logger.getLogger( BodymediaDailySynch.class );
	
	@Resource
	private ProviderDao providerDao;	
	
	@Resource
	private BodymediaService bodymediaService;
	
	@Resource
	private UserDao userDao;
	
	@Resource
	private MeasurementsDao measurementsDao;
	
	@Resource
	private BadgeService badgeService;


	
	@Override
	public ProviderType getProviderType() {
		return ProviderType.BODY_MEDIA;
	}
	
	
	@Transactional
	@Override
	public void synchGoalsAndActivity( String userId ){
		
		 LOG.info("Synchronization with BodyMedia for '" + userId + "' started...");
		 
		 Set<BadgeMeasurement> synchParams = providerService.getSynchForProvider( userId, getProviderType() );
			
		 if( synchParams.isEmpty() ){
			 return;
		 }
		 
		 User user = userDao.getUserByID(userId);
		 
		 ProviderCredentials bodymediaCredentials = providerDao.getProviderCredentials(userId, ProviderType.BODY_MEDIA);
		 
		 if( bodymediaCredentials == null ){
			 LOG.error("Can't find registered credentials for BodyMedia");
			 throw new InternalServerException("Can't find registered credentials for BodyMedia");
		 }
		
		Token accessToken = new Token( bodymediaCredentials.getAccessToken().getToken(), bodymediaCredentials.getAccessToken().getSecret() );
		
		
		// synch weight		
		synchWeight(user, accessToken, synchParams);	
		
		// synch all measurements except "WEIGHT"
		synchMeasurements(user, accessToken, synchParams);
		
		// synch goals		
		synchGoals(user, accessToken, synchParams );
	}

	private void synchMeasurements(User user, Token accessToken, Set<BadgeMeasurement> synchParams ) {
		
		BMMeasurements measureDto = bodymediaService.getMeasurements( accessToken );	
		
		synchSingleMeasurement(user, measureDto.getSleep(), BadgeMeasurement.SLEEP, synchParams);
		synchSingleMeasurement(user, measureDto.getStep(), BadgeMeasurement.STEPS, synchParams);
		synchSingleMeasurement(user, measureDto.getConsumption(), BadgeMeasurement.CONSUMED_CALORIES, synchParams);
		synchSingleMeasurement(user, measureDto.getBurn(), BadgeMeasurement.BURNED_CALORIES, synchParams);
	}

	private void synchWeight(User user, Token accessToken, Set<BadgeMeasurement> synchParams ) {
		
		if( ! synchParams.contains(BadgeMeasurement.WEIGHT) ){
			return;
		}
		
		BMWeightMeasureDto dto = bodymediaService.getWeight( accessToken );

		Measurement measure = new Measurement();
		measure.setDate( dto.getStartDate() );
		measure.setSrcName( ProviderType.BODY_MEDIA.name() );
		measure.setType( BadgeMeasurement.WEIGHT );
		measure.setValue( dto.getValue(), dto.getMeasure() );
		measure.setUser( user );		
		measurementsDao.saveOrUpdate( measure );
		
	}
	
	private void synchSingleMeasurement( User user, BMBaseMeasurementDto dto, BadgeMeasurement type, Set<BadgeMeasurement> synchParams  ){
		
		if( ! synchParams.contains(type) ){
			return;
		}
		
		Measurement measure = new Measurement();
		measure.setDate( dto.getStartDate() );
		measure.setSrcName( ProviderType.BODY_MEDIA.name() );
		measure.setType( type );
		measure.setValue( BigDecimal.valueOf(dto.getValueAsLong()) );
		measure.setUser( user );		
		measurementsDao.saveOrUpdate( measure );
	}


	private void synchGoals(User user, Token accessToken,  Set<BadgeMeasurement> synchParams ) {
		
		BMGoalsDto goalsDto = bodymediaService.getGoals( accessToken );
		
		if( goalsDto == null ){
			LOG.warn("No goals for BodyMedia to synchronise");
			return;
		}
		
		for( BMGoal goalDto : goalsDto.getGoals() ){
			
			BadgeMeasurement goalType = BMGoalTypeFactory.INST.getType( goalDto.getType() );
			
			if( goalType == BadgeMeasurement.UNKNOWN ){
				LOG.warn("Unknown goal type received from BodyMedia: '" + goalDto.getType() + "'");
			}
			else {
				badgeService.updateGoal( ProviderType.BODY_MEDIA.name(), user.getId(), goalType, goalDto.getValue().longValue(), synchParams );
			}
		}
		
	}

}
